from ast import Module
from ast import stmt
from typing import List as ListType

import ast
import inspect
from ast import PyCF_ONLY_AST, PyCF_ALLOW_TOP_LEVEL_AWAIT
import types
import os
from contextlib import contextmanager
PyCF_DONT_IMPLY_DEDENT = 0x200  # Matches pythonrun.h

_assign_nodes = (ast.AugAssign, ast.AnnAssign, ast.Assign)
_single_targets_nodes = (ast.AugAssign, ast.AnnAssign)

user_module = types.ModuleType("__main__",
                               doc="Automatically created module for IPython interactive environment")

stored = []


def tracefunc(frame, event, arg):
    if '_debugger_case_scoped_stepping_target' not in frame.f_code.co_filename:
        return None
    stored.append(frame)
    print('\n---')
    print(event, id(frame), os.path.basename(frame.f_code.co_filename), frame.f_lineno, arg, frame.f_code.co_name)
    assert frame.f_back.f_code.co_name == 'run_code'
    return None


@contextmanager
def tracing_info():
    import sys
    sys.settrace(tracefunc)
    try:
        yield
    finally:
        sys.settrace(None)


# Note: this is roughly what IPython itself does at:
# https://github.com/ipython/ipython/blob/master/IPython/core/interactiveshell.py
class Runner:

    async def run_ast_nodes(
        self,
        nodelist: ListType[stmt],
        cell_name: str,
        interactivity="last_expr",
        compiler=compile,
    ):
        if not nodelist:
            return

        if interactivity == 'last_expr_or_assign':
            if isinstance(nodelist[-1], _assign_nodes):
                asg = nodelist[-1]
                if isinstance(asg, ast.Assign) and len(asg.targets) == 1:
                    target = asg.targets[0]
                elif isinstance(asg, _single_targets_nodes):
                    target = asg.target
                else:
                    target = None
                if isinstance(target, ast.Name):
                    nnode = ast.Expr(ast.Name(target.id, ast.Load()))
                    ast.fix_missing_locations(nnode)
                    nodelist.append(nnode)
            interactivity = 'last_expr'

        _async = False
        if interactivity == 'last_expr':
            if isinstance(nodelist[-1], ast.Expr):
                interactivity = "last"
            else:
                interactivity = "none"

        if interactivity == 'none':
            to_run_exec, to_run_interactive = nodelist, []
        elif interactivity == 'last':
            to_run_exec, to_run_interactive = nodelist[:-1], nodelist[-1:]
        elif interactivity == 'all':
            to_run_exec, to_run_interactive = [], nodelist
        else:
            raise ValueError("Interactivity was %r" % interactivity)

        def compare(code):
            is_async = inspect.CO_COROUTINE & code.co_flags == inspect.CO_COROUTINE
            return is_async

        # Refactor that to just change the mod constructor.
        to_run = []
        for node in to_run_exec:
            to_run.append((node, "exec"))

        for node in to_run_interactive:
            to_run.append((node, "single"))

        for node, mode in to_run:
            if mode == "exec":
                mod = Module([node], [])
            elif mode == "single":
                mod = ast.Interactive([node])
            code = compiler(mod, cell_name, mode, PyCF_DONT_IMPLY_DEDENT |
                            PyCF_ALLOW_TOP_LEVEL_AWAIT)
            asy = compare(code)
            if await self.run_code(code, async_=asy):
                return True

    async def run_code(self, code_obj, *, async_=False):
        if async_:
            await eval(code_obj, self.user_global_ns, self.user_ns)
        else:
            exec(code_obj, self.user_global_ns, self.user_ns)

    @property
    def user_global_ns(self):
        return user_module.__dict__

    @property
    def user_ns(self):
        return user_module.__dict__


async def main():
    SCOPED_STEPPING_TARGET = os.getenv('SCOPED_STEPPING_TARGET', '_debugger_case_scoped_stepping_target.py')
    filename = os.path.join(os.path.dirname(__file__), SCOPED_STEPPING_TARGET)
    assert os.path.exists(filename), '%s does not exist.' % (filename,)
    with open(filename, 'r') as stream:
        source = stream.read()
    code_ast = compile(
        source,
        filename,
        'exec',
        PyCF_DONT_IMPLY_DEDENT | PyCF_ONLY_AST | PyCF_ALLOW_TOP_LEVEL_AWAIT,
        1)

    runner = Runner()
    await runner.run_ast_nodes(code_ast.body, filename)


if __name__ == '__main__':
    import asyncio
    asyncio.run(main())
    print('TEST SUCEEDED!')



# coding: utf-8
from __future__ import unicode_literals
from _pydev_bundle._pydev_completer import isidentifier, extract_token_and_qualifier, TokenAndQualifier


def test_isidentifier():
    assert isidentifier("abc")
    assert not isidentifier("<")
    assert not isidentifier("")
    assert isidentifier("áéíóú")


def test_extract_token_and_qualifier():
    assert extract_token_and_qualifier("tok", 0, 0) == TokenAndQualifier("", "")
    assert extract_token_and_qualifier("tok", 0, 1) == TokenAndQualifier("", "t")
    assert extract_token_and_qualifier("tok", 0, 2) == TokenAndQualifier("", "to")
    assert extract_token_and_qualifier("tok", 0, 3) == TokenAndQualifier("", "tok")
    assert extract_token_and_qualifier("tok", 0, 4) == TokenAndQualifier("", "tok")

    assert extract_token_and_qualifier("tok.qual", 0, 0) == TokenAndQualifier("", "")
    assert extract_token_and_qualifier("tok.qual", 0, 1) == TokenAndQualifier("", "t")
    assert extract_token_and_qualifier("tok.qual", 0, 2) == TokenAndQualifier("", "to")
    assert extract_token_and_qualifier("tok.qual", 0, 3) == TokenAndQualifier("", "tok")

    assert extract_token_and_qualifier("tok.qual", 0, 4) == TokenAndQualifier("tok", "")
    assert extract_token_and_qualifier("tok.qual", 0, 5) == TokenAndQualifier("tok", "q")
    assert extract_token_and_qualifier("tok.qual", 0, 6) == TokenAndQualifier("tok", "qu")
    assert extract_token_and_qualifier("tok.qual", 0, 7) == TokenAndQualifier("tok", "qua")
    assert extract_token_and_qualifier("tok.qual", 0, 8) == TokenAndQualifier("tok", "qual")

    # out of range (column)
    assert extract_token_and_qualifier("tok.qual.qual2", 0, 100) == TokenAndQualifier("tok.qual", "qual2")

    assert extract_token_and_qualifier("t<ok", 0, 0) == TokenAndQualifier("", "")
    assert extract_token_and_qualifier("t<ok", 0, 1) == TokenAndQualifier("", "t")
    assert extract_token_and_qualifier("t<ok", 0, 2) == TokenAndQualifier("", "")
    assert extract_token_and_qualifier("t<ok", 0, 3) == TokenAndQualifier("", "o")
    assert extract_token_and_qualifier("t<ok", 0, 4) == TokenAndQualifier("", "ok")

    assert extract_token_and_qualifier("a\nt<ok", 1, 0) == TokenAndQualifier("", "")
    assert extract_token_and_qualifier("a\nt<ok", 1, 1) == TokenAndQualifier("", "t")
    assert extract_token_and_qualifier("a\nt<ok", 1, 2) == TokenAndQualifier("", "")
    assert extract_token_and_qualifier("a\nt<ok", 1, 3) == TokenAndQualifier("", "o")
    assert extract_token_and_qualifier("a\nt<ok", 1, 4) == TokenAndQualifier("", "ok")

    # out of range (line)
    assert extract_token_and_qualifier("a\nt<ok", 5, 4) == TokenAndQualifier("", "")



import os
import pytest

import time

from contextlib import contextmanager
from tests_python.debugger_unittest import IS_CPYTHON
from tests_python.debug_constants import TEST_CYTHON
from _pydevd_bundle.pydevd_constants import IS_PY312_OR_GREATER, IS_PY311_OR_GREATER, IS_PY313_OR_GREATER, TODO_PY313_OR_GREATER

pytest_plugins = [
    str("tests_python.debugger_fixtures"),
]

pytestmark = pytest.mark.skipif(not IS_CPYTHON or not TEST_CYTHON, reason="Requires binaries")


@pytest.fixture
def case_setup_force_frame_eval(case_setup):
    def get_environ(writer):
        env = os.environ.copy()
        if IS_PY311_OR_GREATER:
            env["PYDEVD_USE_FRAME_EVAL"] = "YES"
        env["PYDEVD_USE_CYTHON"] = "YES"
        return env

    original_test_file = case_setup.test_file

    @contextmanager
    def test_file(*args, **kwargs):
        kwargs.setdefault("get_environ", get_environ)
        with original_test_file(*args, **kwargs) as writer:
            yield writer

    case_setup.test_file = test_file
    return case_setup


def test_step_and_resume(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file("_debugger_case_simple_calls.py") as writer:
        writer.write_add_breakpoint(10, "Method2")
        writer.write_add_breakpoint(2, "Method1")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit()
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        elif IS_PY311_OR_GREATER:
            assert hit.suspend_type == "trace"
        else:
            assert hit.suspend_type == "frame_eval"
        assert hit.line == 10

        writer.write_step_over(hit.thread_id)
        hit = writer.wait_for_breakpoint_hit("108")

        assert hit.line == 11
        # we use tracing debugger while stepping
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        else:
            assert hit.suspend_type == "trace"

        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit()

        assert hit.line == 2
        # we enable frame evaluation debugger after "Resume" command
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        elif IS_PY311_OR_GREATER:
            assert hit.suspend_type == "trace"
        else:
            assert hit.suspend_type == "frame_eval"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_step_return(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file("_debugger_case56.py") as writer:
        writer.write_add_breakpoint(2, "Call2")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit()

        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        elif IS_PY311_OR_GREATER:
            assert hit.suspend_type == "trace"
        else:
            assert hit.suspend_type == "frame_eval"
        writer.write_get_frame(hit.thread_id, hit.frame_id)

        writer.write_step_return(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit("109")

        assert hit.line == 8
        # Step return uses temporary breakpoint, so we use tracing debugger
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        else:
            assert hit.suspend_type == "trace"

        writer.write_step_in(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit("107")

        # goes to line 4 in jython (function declaration line)
        assert hit.line in (4, 5)
        # we use tracing debugger for stepping
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        else:
            assert hit.suspend_type == "trace", 'Expected suspend type to be "trace", but was: %s' % hit.suspend_type

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_add_break_while_running(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file("_debugger_case3.py") as writer:
        writer.write_make_initial_run()
        time.sleep(0.5)
        breakpoint_id = writer.write_add_breakpoint(4, "")

        hit = writer.wait_for_breakpoint_hit()

        assert hit.line == 4
        # we use tracing debugger if breakpoint was added while running
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        else:
            assert hit.suspend_type == "trace"

        writer.write_get_frame(hit.thread_id, hit.frame_id)

        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit()
        assert hit.line == 4
        # we still use tracing debugger
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        else:
            assert hit.suspend_type == "trace"

        writer.write_get_frame(hit.thread_id, hit.frame_id)

        writer.write_remove_breakpoint(breakpoint_id)

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_exc_break(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file("_debugger_case_simple_calls.py") as writer:
        writer.write_add_breakpoint(10, "Method2")
        writer.write_add_exception_breakpoint_with_policy("IndexError", "1", "0", "0")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit()

        assert hit.line == 10
        # we use tracing debugger if there are exception breakpoints
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        else:
            assert hit.suspend_type == "trace"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_add_exc_break_while_running(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file("_debugger_case_simple_calls.py") as writer:
        writer.write_add_breakpoint(10, "Method2")
        writer.write_add_breakpoint(2, "Method1")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit("111")

        assert hit.line == 10
        # we use tracing debugger if there are exception breakpoints
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        elif IS_PY311_OR_GREATER:
            assert hit.suspend_type == "trace"
        else:
            assert hit.suspend_type == "frame_eval"

        writer.write_add_exception_breakpoint_with_policy("IndexError", "1", "0", "0")

        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit()

        assert hit.line == 2
        # we use tracing debugger if exception break was added
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        else:
            assert hit.suspend_type == "trace"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_add_termination_exc_break(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file("_debugger_case_simple_calls.py") as writer:
        writer.write_add_breakpoint(10, "Method2")
        writer.write_add_exception_breakpoint_with_policy("IndexError", "0", "1", "0")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit(line=10)

        # we can use frame evaluation with exception breakpoint with "On termination" suspend policy
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        elif IS_PY311_OR_GREATER:
            assert hit.suspend_type == "trace"
        else:
            assert hit.suspend_type == "frame_eval"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


@pytest.mark.skipif(IS_PY311_OR_GREATER, reason="frame eval not supported in python 3.11 or newer.")
def test_frame_eval_whitebox_test(case_setup_force_frame_eval):
    from tests_python.debugger_unittest import CMD_STEP_INTO, CMD_STEP_RETURN, CMD_STEP_OVER

    with case_setup_force_frame_eval.test_file("_debugger_case_frame_eval.py") as writer:
        line_on_global = writer.get_line_index_with_content("break on global")
        writer.write_add_breakpoint(line_on_global, "")
        writer.write_add_breakpoint(writer.get_line_index_with_content("break on check_with_trace"), "None")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit(line=line_on_global)
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        elif IS_PY311_OR_GREATER:
            assert hit.suspend_type == "trace"
        else:
            assert hit.suspend_type == "frame_eval"

        writer.write_step_over(hit.thread_id)
        hit = writer.wait_for_breakpoint_hit(reason=CMD_STEP_OVER)

        writer.write_step_in(hit.thread_id)
        hit = writer.wait_for_breakpoint_hit(
            line=writer.get_line_index_with_content("check_step_in_then_step_return") + 1, reason=CMD_STEP_INTO
        )
        assert hit.name == "check_step_in_then_step_return"

        writer.write_step_return(hit.thread_id)
        hit = writer.wait_for_breakpoint_hit(reason=CMD_STEP_RETURN)
        assert hit.name == "<module>"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_frame_eval_change_breakpoints(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file("_debugger_case_change_breaks.py") as writer:
        break1_line = writer.get_line_index_with_content("break 1")
        break2_line = writer.get_line_index_with_content("break 2")

        break2_id = writer.write_add_breakpoint(break2_line, "None")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit(line=break2_line)
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        elif IS_PY311_OR_GREATER:
            assert hit.suspend_type == "trace"
        else:
            assert hit.suspend_type == "frame_eval"

        writer.write_remove_breakpoint(break2_id)
        writer.write_add_breakpoint(break1_line, "None")
        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit(line=break1_line)
        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


@pytest.mark.skipif(TODO_PY313_OR_GREATER, reason="Flakey on python 3.13 or newer")
def test_generator_code_cache(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file("_debugger_case_yield_from.py") as writer:
        break1_line = writer.get_line_index_with_content("break1")
        writer.write_add_breakpoint(break1_line)
        break2_line = writer.get_line_index_with_content("break2")
        writer.write_add_breakpoint(break2_line)
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit(line=break1_line)
        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit(line=break2_line)
        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit(line=break2_line)
        writer.write_run_thread(hit.thread_id)

        hit = writer.wait_for_breakpoint_hit(line=break2_line)
        writer.write_run_thread(hit.thread_id)

        if not IS_PY312_OR_GREATER:
            # On python 3.12 there's no list-comprehension additional frame,
            # so, one less step is expected.
            hit = writer.wait_for_breakpoint_hit(line=break2_line)
            writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True


def test_break_line_1(case_setup_force_frame_eval):
    with case_setup_force_frame_eval.test_file("_debugger_case_yield_from.py") as writer:
        break1_line = 1
        break1_id = writer.write_add_breakpoint(break1_line, "None")
        writer.write_make_initial_run()

        hit = writer.wait_for_breakpoint_hit(line=break1_line)
        if IS_PY312_OR_GREATER:
            assert hit.suspend_type == "sys_monitor"
        elif IS_PY311_OR_GREATER:
            assert hit.suspend_type == "trace"
        else:
            assert hit.suspend_type == "frame_eval"

        writer.write_run_thread(hit.thread_id)

        writer.finished_ok = True



from _pydevd_bundle._debug_adapter.pydevd_schema import InitializeRequest, InitializeRequestArguments, InitializeResponse, Capabilities
from _pydevd_bundle._debug_adapter import pydevd_schema, pydevd_base_schema
from _pydevd_bundle._debug_adapter.pydevd_schema import ThreadsResponse


def test_schema():
    json_msg = """
{
    "arguments": {
        "adapterID": "pydevd",
        "clientID": "vscode",
        "clientName": "Visual Studio Code",
        "columnsStartAt1": true,
        "linesStartAt1": true,
        "locale": "en-us",
        "pathFormat": "path",
        "supportsRunInTerminalRequest": true,
        "supportsVariablePaging": true,
        "supportsVariableType": true
    },
    "command": "initialize",
    "seq": 1,
    "type": "request"
}"""

    initialize_request = pydevd_base_schema.from_json(json_msg)
    assert initialize_request.__class__ == InitializeRequest
    assert initialize_request.arguments.__class__ == InitializeRequestArguments
    assert initialize_request.arguments.adapterID == "pydevd"
    assert initialize_request.command == "initialize"
    assert initialize_request.type == "request"
    assert initialize_request.seq == 1

    response = pydevd_base_schema.build_response(initialize_request)
    assert response.__class__ == InitializeResponse
    assert response.seq == -1  # Must be set before sending
    assert response.command == "initialize"
    assert response.type == "response"
    assert response.body.__class__ == Capabilities

    assert response.to_dict() == {"seq": -1, "type": "response", "request_seq": 1, "success": True, "command": "initialize", "body": {}}

    capabilities = response.body  # : :type capabilities: Capabilities
    capabilities.supportsCompletionsRequest = True
    assert response.to_dict() == {
        "seq": -1,
        "type": "response",
        "request_seq": 1,
        "success": True,
        "command": "initialize",
        "body": {"supportsCompletionsRequest": True},
    }

    initialize_event = pydevd_schema.InitializedEvent()
    assert initialize_event.to_dict() == {"seq": -1, "type": "event", "event": "initialized"}


def test_schema_translation_frame():
    pydevd_base_schema.BaseSchema.initialize_ids_translation()
    stack_trace_arguments = pydevd_schema.StackTraceArguments(threadId=1)
    stack_trace_request = pydevd_schema.StackTraceRequest(stack_trace_arguments)

    stackFrames = [
        pydevd_schema.StackFrame(id=2**45, name="foo", line=1, column=1).to_dict(),
        pydevd_schema.StackFrame(id=2**46, name="bar", line=1, column=1).to_dict(),
    ]
    body = pydevd_schema.StackTraceResponseBody(stackFrames)
    stack_trace_response = pydevd_base_schema.build_response(stack_trace_request, kwargs=dict(body=body))
    as_dict = stack_trace_response.to_dict(update_ids_to_dap=True)
    assert as_dict == {
        "type": "response",
        "request_seq": -1,
        "success": True,
        "command": "stackTrace",
        "body": {
            "stackFrames": [
                {"id": 1, "name": "foo", "line": 1, "column": 1, "source": {}},
                {"id": 2, "name": "bar", "line": 1, "column": 1, "source": {}},
            ]
        },
        "seq": -1,
    }

    reconstructed = pydevd_base_schema.from_dict(as_dict, update_ids_from_dap=True)
    assert reconstructed.to_dict() == {
        "type": "response",
        "request_seq": -1,
        "success": True,
        "command": "stackTrace",
        "body": {
            "stackFrames": [
                {"id": 2**45, "name": "foo", "line": 1, "column": 1, "source": {}},
                {"id": 2**46, "name": "bar", "line": 1, "column": 1, "source": {}},
            ]
        },
        "seq": -1,
    }


def test_schema_translation_thread():
    from _pydevd_bundle._debug_adapter.pydevd_schema import ThreadsRequest

    pydevd_base_schema.BaseSchema.initialize_ids_translation()

    threads = [
        pydevd_schema.Thread(id=2**45, name="foo").to_dict(),
        pydevd_schema.Thread(id=2**46, name="bar").to_dict(),
    ]
    body = pydevd_schema.ThreadsResponseBody(threads)
    threads_request = ThreadsRequest()
    threads_response = pydevd_base_schema.build_response(threads_request, kwargs=dict(body=body))
    as_dict = threads_response.to_dict(update_ids_to_dap=True)
    assert as_dict == {
        "type": "response",
        "request_seq": -1,
        "success": True,
        "command": "threads",
        "body": {
            "threads": [
                {"id": 1, "name": "foo"},
                {"id": 2, "name": "bar"},
            ]
        },
        "seq": -1,
    }

    reconstructed = pydevd_base_schema.from_dict(as_dict, update_ids_from_dap=True)
    assert reconstructed.to_dict() == {
        "type": "response",
        "request_seq": -1,
        "success": True,
        "command": "threads",
        "body": {"threads": [{"id": 2**45, "name": "foo"}, {"id": 2**46, "name": "bar"}]},
        "seq": -1,
    }



import json

from tests_python.debugger_unittest import ReaderThread, IS_JYTHON
import pytest
import socket
from _pydev_bundle import pydev_localhost
from _pydevd_bundle.pydevd_comm import start_client

ABORT_CONNECTION = "ABORT_CONNECTION"

pytestmark = pytest.mark.skipif(IS_JYTHON, reason="Getting the actual port does not work when the port is == 0.")


class _DummySocket(object):
    def __init__(self):
        self._sock_for_reader_thread = None
        self._sock_for_fixture_test = None
        self._socket_server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._localhost = host = pydev_localhost.get_localhost()
        self._socket_server.bind((host, 0))
        self._socket_server.listen(1)

    def recv(self, *args, **kwargs):
        if self._sock_for_reader_thread is None:
            sock, _addr = self._socket_server.accept()
            self._sock_for_reader_thread = sock
        return self._sock_for_reader_thread.recv(*args, **kwargs)

    def put(self, msg):
        if not isinstance(msg, bytes):
            msg = msg.encode("utf-8")

        if self._sock_for_fixture_test is None:
            self._sock_for_fixture_test = start_client(*self._socket_server.getsockname())

        self._sock_for_fixture_test.sendall(msg)

    def close(self):
        self._socket_server.close()

        if self._sock_for_fixture_test is not None:
            self._sock_for_fixture_test.close()

        if self._sock_for_reader_thread is not None:
            self._sock_for_reader_thread.close()


@pytest.yield_fixture
def _dummy_socket():
    sock = _DummySocket()
    yield sock
    sock.close()


def test_fixture_reader_thread1(_dummy_socket):
    sock = _dummy_socket

    reader_thread = ReaderThread(sock)
    reader_thread.start()

    json_part = json.dumps({"key": "val"})
    json_part = json_part.replace(":", ":\n")
    msg = json_part

    msg = ("Content-Length: %s\r\n\r\n%s" % (len(msg), msg)).encode("utf-8")
    # Check that receiving 2 messages at a time we're able to properly deal
    # with each one.
    sock.put(msg + msg)

    assert reader_thread.get_next_message("check 1") == json_part
    assert reader_thread.get_next_message("check 2") == json_part


def test_fixture_reader_thread2(_dummy_socket):
    sock = _DummySocket()

    reader_thread = ReaderThread(sock)
    reader_thread.start()

    json_part = json.dumps({"key": "val"})
    json_part = json_part.replace(":", ":\n")
    msg = json_part

    http = "Content-Length: %s\r\n\r\n%s" % (len(msg), msg)
    sock.put("msg1\nmsg2\nmsg3\n" + http + http)

    assert reader_thread.get_next_message("check 1") == "msg1"
    assert reader_thread.get_next_message("check 2") == "msg2"
    assert reader_thread.get_next_message("check 3") == "msg3"
    assert reader_thread.get_next_message("check 4") == json_part
    assert reader_thread.get_next_message("check 5") == json_part


def test_fixture_reader_thread3(_dummy_socket):
    sock = _DummySocket()

    reader_thread = ReaderThread(sock)
    reader_thread.start()

    msg = "aaaaaaabbbbbbbccccccc"
    http = "Content-Length: %s\r\n\r\n%s" % (len(msg), msg)
    http *= 2
    initial = http
    for i in range(1, len(http)):
        while http:
            sock.put(http[:i])
            http = http[i:]

        assert reader_thread.get_next_message("check 1: %s" % i) == msg
        assert reader_thread.get_next_message("check 2: %s" % i) == msg
        http = initial



from _pydevd_bundle.pydevd_constants import IS_WINDOWS, IS_MAC


def test_in_project_roots_prefix_01(tmpdir):
    from _pydevd_bundle.pydevd_filtering import FilesFiltering

    files_filtering = FilesFiltering()

    another = str(tmpdir.join("another"))
    assert not another.endswith("/") and not another.endswith("\\")

    files_filtering.set_library_roots([another])
    files_filtering.set_project_roots([])
    assert not files_filtering.in_project_roots(another + "/f.py")
    if IS_WINDOWS:
        assert not files_filtering.in_project_roots(another + "\\f.py")
    else:
        assert files_filtering.in_project_roots(another + "\\f.py")

    assert files_filtering.in_project_roots(another + "f.py")


def test_in_project_roots_prefix_02(tmpdir):
    from _pydevd_bundle.pydevd_filtering import FilesFiltering

    files_filtering = FilesFiltering()

    another = str(tmpdir.join("another"))
    assert not another.endswith("/") and not another.endswith("\\")

    files_filtering.set_library_roots([])
    files_filtering.set_project_roots([another])
    assert files_filtering.in_project_roots(another + "/f.py")
    if IS_WINDOWS:
        assert files_filtering.in_project_roots(another + "\\f.py")
    else:
        assert not files_filtering.in_project_roots(another + "\\f.py")

    assert not files_filtering.in_project_roots(another + "f.py")


def test_in_project_roots(tmpdir):
    from _pydevd_bundle.pydevd_filtering import FilesFiltering

    files_filtering = FilesFiltering()

    import os.path
    import sys

    if IS_WINDOWS:
        assert files_filtering._get_library_roots() == [os.path.normcase(x) + "\\" for x in files_filtering._get_default_library_roots()]
    elif IS_MAC:
        assert files_filtering._get_library_roots() == [x.lower() + "/" for x in files_filtering._get_default_library_roots()]
    else:
        assert files_filtering._get_library_roots() == [os.path.normcase(x) + "/" for x in files_filtering._get_default_library_roots()]

    site_packages = tmpdir.mkdir("site-packages")
    project_dir = tmpdir.mkdir("project")

    project_dir_inside_site_packages = str(site_packages.mkdir("project"))
    site_packages_inside_project_dir = str(project_dir.mkdir("site-packages"))

    # Convert from pytest paths to str.
    site_packages = str(site_packages)
    project_dir = str(project_dir)
    tmpdir = str(tmpdir)

    # Test permutations of project dir inside site packages and vice-versa.
    files_filtering.set_project_roots([project_dir, project_dir_inside_site_packages])
    files_filtering.set_library_roots([site_packages, site_packages_inside_project_dir])

    check = [
        (tmpdir, False),
        (site_packages, False),
        (site_packages_inside_project_dir, False),
        (project_dir, True),
        (project_dir_inside_site_packages, True),
    ]
    for check_path, find in check[:]:
        filename_inside = os.path.join(check_path, "a.py")
        with open(filename_inside, "w") as stream:
            # Note: on Github actions, tmpdir may be something as:
            # c:\\users\\runner~1\\appdata\\local\\temp\\pytest-of-runneradmin\\pytest-0\\test_in_project_roots0
            # internally this may be set as:
            # c:\\users\\runneradmin\\appdata\\local\\temp\\pytest-of-runneradmin\\pytest-0\\test_in_project_roots0
            # So, when getting the absolute path, `runner~1` will be properly expanded to `runneradmin` if the
            # file exists, but if it doesn't it's not (which may make the test fail), so, make sure
            # that we actually create the file so that things work as expected.
            stream.write("...")
        check.append((filename_inside, find))

    for check_path, find in check:
        if files_filtering.in_project_roots(check_path) != find:
            if find:
                msg = "Expected %s to be in the project roots.\nProject roots: %s\nLibrary roots: %s\n"
            else:
                msg = "Expected %s NOT to be in the project roots.\nProject roots: %s\nLibrary roots: %s\n"

            raise AssertionError(
                msg
                % (
                    check_path,
                    files_filtering._get_project_roots(),
                    files_filtering._get_library_roots(),
                )
            )

    files_filtering.set_project_roots([])
    files_filtering.set_library_roots([site_packages, site_packages_inside_project_dir])

    # If the IDE did not set the project roots, consider anything not in the site
    # packages as being in a project root (i.e.: we can calculate default values for
    # site-packages but not for project roots).
    check = [
        (tmpdir, True),
        (site_packages, False),
        (site_packages_inside_project_dir, False),
        (project_dir, True),
        (project_dir_inside_site_packages, False),
        ("<foo>", False),
        ("<ipython>", True),
        ("<frozen importlib._bootstrap>", False),
    ]

    for check_path, find in check:
        assert files_filtering.in_project_roots(check_path) == find, "Expected: %s to be a part of the project: %s" % (check_path, find)

    sys.path.append(str(site_packages))
    try:
        default_library_roots = files_filtering._get_default_library_roots()
        assert len(set(default_library_roots)) == len(default_library_roots), "Duplicated library roots found in: %s" % (
            default_library_roots,
        )

        assert str(site_packages) in default_library_roots
        for path in sys.path:
            if os.path.exists(path) and path.endswith("site-packages"):
                assert path in default_library_roots
    finally:
        sys.path.remove(str(site_packages))


def test_filtering(tmpdir):
    from _pydevd_bundle.pydevd_filtering import FilesFiltering
    from _pydevd_bundle.pydevd_filtering import ExcludeFilter

    files_filtering = FilesFiltering()

    site_packages = tmpdir.mkdir("site-packages")
    project_dir = tmpdir.mkdir("project")

    project_dir_inside_site_packages = str(site_packages.mkdir("project"))
    site_packages_inside_project_dir = str(project_dir.mkdir("site-packages"))

    files_filtering.set_exclude_filters(
        [
            ExcludeFilter("**/project*", True, True),
            ExcludeFilter("**/bar*", False, True),
        ]
    )
    assert files_filtering.exclude_by_filter("/foo/project", None) is True
    assert files_filtering.exclude_by_filter("/foo/unmatched", None) is None
    assert files_filtering.exclude_by_filter("/foo/bar", None) is False


def test_glob_matching():
    from _pydevd_bundle.pydevd_filtering import glob_matches_path

    # Linux
    for sep, altsep in (("\\", "/"), ("/", None)):

        def build(path):
            if sep == "/":
                return path
            else:
                return ("c:" + path).replace("/", "\\")

        assert glob_matches_path(build("/a"), r"*", sep, altsep)

        assert not glob_matches_path(build("/a/b/c/some.py"), "/a/**/c/so?.py", sep, altsep)

        assert glob_matches_path("/a/b/c", "/a/b/*")
        assert not glob_matches_path("/a/b", "/*")
        assert glob_matches_path("/a/b", "/*/b")
        assert glob_matches_path("/a/b", "**/*")
        assert not glob_matches_path("/a/b", "**/a")

        assert glob_matches_path(build("/a/b/c/d"), "**/d", sep, altsep)
        assert not glob_matches_path(build("/a/b/c/d"), "**/c", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d"), "**/c/d", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d"), "**/b/c/d", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d"), "/*/b/*/d", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d"), "**/c/*", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d"), "/a/**/c/*", sep, altsep)

        # I.e. directories are expected to end with '/', so, it'll match
        # something as **/directory/**
        assert glob_matches_path(build("/a/b/c/"), "**/c/**", sep, altsep)
        assert glob_matches_path(build("/a/b/c/"), "**/c/", sep, altsep)
        # But not something as **/directory (that'd be a file match).
        assert not glob_matches_path(build("/a/b/c/"), "**/c", sep, altsep)
        assert not glob_matches_path(build("/a/b/c"), "**/c/", sep, altsep)

        assert glob_matches_path(build("/a/b/c/d.py"), "/a/**/c/*", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d.py"), "/a/**/c/*.py", sep, altsep)
        assert glob_matches_path(build("/a/b/c/some.py"), "/a/**/c/so*.py", sep, altsep)
        assert glob_matches_path(build("/a/b/c/some.py"), "/a/**/c/som?.py", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d"), "/**", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d"), "/**/d", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d.py"), "/**/*.py", sep, altsep)
        assert glob_matches_path(build("/a/b/c/d.py"), "**/c/*.py", sep, altsep)

        if IS_WINDOWS:
            assert glob_matches_path(build("/a/b/c/d.py"), "**/C/*.py", sep, altsep)
            assert glob_matches_path(build("/a/b/C/d.py"), "**/c/*.py", sep, altsep)

        # Expected not to match.
        assert not glob_matches_path(build("/a/b/c/d"), "/**/d.py", sep, altsep)
        assert not glob_matches_path(build("/a/b/c/d.pyx"), "/a/**/c/*.py", sep, altsep)
        assert not glob_matches_path(build("/a/b/c/d"), "/*/d", sep, altsep)

        if sep == "/":
            assert not glob_matches_path(build("/a/b/c/d"), r"**\d", sep, altsep)  # Match with \ doesn't work on linux...
            assert not glob_matches_path(build("/a/b/c/d"), r"c:\**\d", sep, altsep)  # Match with drive doesn't work on linux...
        else:
            # Works in Windows.
            assert glob_matches_path(build("/a/b/c/d"), r"**\d", sep, altsep)
            assert glob_matches_path(build("/a/b/c/d"), r"c:\**\d", sep, altsep)

        # Corner cases
        assert not glob_matches_path(build("/"), r"", sep, altsep)
        assert glob_matches_path(build(""), r"", sep, altsep)
        assert not glob_matches_path(build(""), r"**", sep, altsep)
        assert glob_matches_path(build("/"), r"**", sep, altsep)
        assert glob_matches_path(build("/"), r"*", sep, altsep)


def test_rules_to_exclude_filter(tmpdir):
    from _pydevd_bundle.pydevd_process_net_command_json import _convert_rules_to_exclude_filters
    from _pydevd_bundle.pydevd_filtering import ExcludeFilter
    from random import shuffle

    dira = tmpdir.mkdir("a")
    dirb = dira.mkdir("b")
    fileb = dirb.join("fileb.py")
    fileb2 = dirb.join("fileb2.py")
    with fileb.open("w") as stream:
        stream.write("")

    def on_error(msg):
        raise AssertionError(msg)

    rules = [
        {"path": str(dira), "include": False},
        {"path": str(dirb), "include": True},
        {"path": str(fileb), "include": True},
        {"path": str(fileb2), "include": True},
        {"path": "**/foo/*.py", "include": True},
        {"module": "bar", "include": False},
        {"module": "bar.foo", "include": True},
    ]
    shuffle(rules)
    exclude_filters = _convert_rules_to_exclude_filters(rules, on_error)
    assert exclude_filters == [
        ExcludeFilter(name=str(fileb2), exclude=False, is_path=True),
        ExcludeFilter(name=str(fileb), exclude=False, is_path=True),
        ExcludeFilter(name=str(dirb) + "/**", exclude=False, is_path=True),
        ExcludeFilter(name=str(dira) + "/**", exclude=True, is_path=True),
        ExcludeFilter(name="**/foo/*.py", exclude=False, is_path=True),
        ExcludeFilter(name="bar.foo", exclude=False, is_path=False),
        ExcludeFilter(name="bar", exclude=True, is_path=False),
    ]



import asyncio
import sys

# Useful for debugging function breakpoint:
# from _pydevd_bundle.pydevd_breakpoints import FunctionBreakpoint
# from _pydevd_bundle.pydevd_api import PyDevdAPI
# from _pydevd_bundle.pydevd_constants import get_global_debugger
#
# function_breakpoints = []
# function_breakpoints.append(
#     FunctionBreakpoint('gen', condition=None, expression=None, suspend_policy='ALL', hit_condition=None, is_logpoint=False))
#
# py_db = get_global_debugger()
# PyDevdAPI().set_function_breakpoints(py_db, function_breakpoints)


async def gen():
    for i in range(10):
        yield i


async def run():
    async for p in gen():
        print(p)


if __name__ == "__main__":
    if sys.version_info[:2] >= (3, 11):
        asyncio.run(run())
    else:
        loop = asyncio.get_event_loop_policy().get_event_loop()
        loop.run_until_complete(run())
    print('TEST SUCEEDED')



import sys
import time

exit_code = eval(sys.argv[1])
print("sys.exit(%r)" % (exit_code,))
print('TEST SUCEEDED!')
try:
    sys.exit(exit_code)  # @handled
except SystemExit:
    pass
sys.exit(exit_code)  # @unhandled





from io import StringIO
import os
import sys
import traceback

import pytest

from tests_python.debug_constants import IS_PY36_OR_GREATER, IS_CPYTHON, TEST_CYTHON, IS_PY311_OR_GREATER
import dis

pytestmark = pytest.mark.skipif(
    not IS_PY36_OR_GREATER or IS_PY311_OR_GREATER or not IS_CPYTHON or not TEST_CYTHON, reason="Requires CPython >= 3.6 < 3.11"
)


class _Tracer(object):
    def __init__(self):
        self.stream = StringIO()
        self._in_print = False
        self.accept_frame = None  # Can be set to a callable
        self.lines_executed = set()

    def tracer_printer(self, frame, event, arg):
        if self._in_print:
            return None
        self._in_print = True
        try:
            if self.accept_frame is None or self.accept_frame(frame):
                if arg is not None:
                    if event == "exception":
                        arg = arg[0].__name__
                    elif arg is not None:
                        arg = str(arg)
                if arg is None:
                    arg = ""

                self.lines_executed.add(frame.f_lineno)

                s = " ".join(
                    (
                        str(frame.f_lineno),
                        frame.f_code.co_name,
                        os.path.basename(frame.f_code.co_filename),
                        event.upper() if event != "line" else event,
                        arg,
                    )
                )
                self.writeln(s)
        except:
            traceback.print_exc()
        self._in_print = False
        return self.tracer_printer

    def writeln(self, s):
        self.write(s)
        self.write("\n")

    def write(self, s):
        if isinstance(s, bytes):
            s = s.decode("utf-8")
        self.stream.write(s)

    def call(self, c):
        sys.settrace(self.tracer_printer)
        c()
        sys.settrace(None)
        return self.stream.getvalue()


def check(
    filename,
    method,
    method_kwargs=None,
    skip_breaks_at_lines=None,
    method_to_change=None,
    stop_at_all_lines=False,
    has_line_event_optimized_in_original_case=False,
):
    """
    :param has_line_event_optimized_in_original_case:
        If True, we're handling a case where we have a double jump, i.e.: some case
        where there's a JUMP_FORWARD which points to a JUMP_ABSOLUTE and this is
        optimized so that the JUMP_FORWARD is changed directly to a JUMP_ABSOLUTE and
        we end up skipping one line event which is supposed to be there but isn't in
        the initial case but appears when we run after modifying the bytecode in memory.

        See: https://github.com/microsoft/debugpy/issues/973#issuecomment-1178090731
    """
    from _pydevd_frame_eval.pydevd_modify_bytecode import _get_code_line_info
    from _pydevd_frame_eval import pydevd_modify_bytecode

    if method_to_change is None:
        method_to_change = method

    if method_kwargs is None:
        method_kwargs = {}
    if skip_breaks_at_lines is None:
        skip_breaks_at_lines = set()

    pydev_break_stops = []

    def _pydev_needs_stop_at_break(line):
        pydev_break_stops.append(line)
        return False

    tracer = _Tracer()

    def accept_frame(f):
        return filename in f.f_code.co_filename

    code = method_to_change.__code__
    code_line_info = _get_code_line_info(code)

    try:
        tracer.accept_frame = accept_frame

        def call():
            method(**method_kwargs)

        tracer.call(call)
        breakpoint_hit_at_least_once = False

        # Ok, we just ran the tracer once without any breakpoints.
        #
        # Gather its tracing profile: this will be our baseline for further tests (it should contain
        # the events and the order in which the were executed).
        #
        # Note: methods cannot have random elements when executing (otherwise
        # the order would be different and the test would be expected to fail).
        baseline = tracer.stream.getvalue()

        for line in sorted(code_line_info.line_to_offset):
            if line in skip_breaks_at_lines:
                continue
            # Now, for each valid line, add a breakpoint and check if the tracing profile is exactly
            # the same (and if the line where we added the breakpoint was executed, see if our
            # callback got called).
            success, new_code = pydevd_modify_bytecode.insert_pydevd_breaks(
                code, set([line]), _pydev_needs_stop_at_break=_pydev_needs_stop_at_break
            )

            assert success
            method_to_change.__code__ = new_code

            tracer = _Tracer()
            tracer.accept_frame = accept_frame
            tracer.call(call)
            contents = tracer.stream.getvalue()

            assert tracer.lines_executed
            if has_line_event_optimized_in_original_case:
                lines = sorted(set(x[1] for x in dis.findlinestarts(new_code) if x[1] is not None))
                new_line_contents = []
                last_line = str(max(lines)) + " "
                for l in contents.splitlines(keepends=True):
                    if not l.strip().startswith(last_line):
                        new_line_contents.append(l)
                contents = "".join(new_line_contents)

            if line in tracer.lines_executed:
                assert set([line]) == set(pydev_break_stops)
                breakpoint_hit_at_least_once = True
            else:
                if stop_at_all_lines:
                    raise AssertionError("Expected the debugger to stop at all lines. Did not stop at line: %s" % (line,))
            del pydev_break_stops[:]

            if baseline != contents:
                print("------- replacement at line: %s ---------" % (line,))
                print("------- baseline ---------")
                print(baseline)
                print("------- contents ---------")
                print(contents)
                print("-------- error -----------")
                assert baseline == contents

        # We must have found a break at least once!
        assert breakpoint_hit_at_least_once
    finally:
        method_to_change.__code__ = code


def test_set_pydevd_break_01():
    from tests_python.resources import _bytecode_overflow_example

    check(
        "_bytecode_overflow_example.py",
        _bytecode_overflow_example.Dummy.fun,
        method_kwargs={"text": "ing"},
        has_line_event_optimized_in_original_case=True,
    )


def test_set_pydevd_break_01a():
    from tests_python.resources import _bytecode_overflow_example

    check("_bytecode_overflow_example.py", _bytecode_overflow_example.check_backtrack, method_kwargs={"x": "f"})


def test_set_pydevd_break_02():
    from tests_python.resources import _bytecode_many_names_example

    check("_bytecode_many_names_example.py", _bytecode_many_names_example.foo)


def test_set_pydevd_break_03():
    from tests_python.resources import _bytecode_big_method

    check("_bytecode_big_method.py", _bytecode_big_method.foo)


def test_set_pydevd_break_04():
    from tests_python.resources import _debugger_case_yield_from

    check("_debugger_case_yield_from.py", _debugger_case_yield_from.method)


def test_set_pydevd_break_05():
    from tests_python import debugger_unittest

    add_to_pythonpath = debugger_unittest._get_debugger_test_file("wrong_bytecode")
    sys.path.append(add_to_pythonpath)

    try:
        with open(debugger_unittest._get_debugger_test_file("wrong_bytecode/_debugger_case_wrong_bytecode.py"), "r") as stream:
            contents = stream.read()

        code = compile(contents, "_my_file_debugger_case_wrong_bytecode.py", "exec")

        def method():
            pass

        method.__code__ = code

        check("_my_file_debugger_case_wrong_bytecode.py", method, skip_breaks_at_lines=set([1]))
    finally:
        sys.path.remove(add_to_pythonpath)


def test_set_pydevd_break_06(pyfile):
    from tests_python.resources import _bytecode_super

    check("_bytecode_super.py", _bytecode_super.B, method_to_change=_bytecode_super.B.__init__, stop_at_all_lines=True)


def test_set_pydevd_break_07():
    from tests_python.resources import _bytecode_overflow_example

    check("_bytecode_overflow_example.py", _bytecode_overflow_example.offset_overflow, method_kwargs={"stream": StringIO()})


def test_set_pydevd_break_08():
    from tests_python.resources import _bytecode_overflow_example

    check("_bytecode_overflow_example.py", _bytecode_overflow_example.long_lines, stop_at_all_lines=True)


def test_internal_double_linked_list():
    from _pydevd_frame_eval.pydevd_modify_bytecode import _HelperBytecodeList

    lst = _HelperBytecodeList()
    node1 = lst.append(1)
    assert list(lst) == [1]
    node2 = lst.append(2)
    assert list(lst) == [1, 2]

    node15 = node1.append(1.5)
    assert list(lst) == [1, 1.5, 2]

    node12 = node15.prepend(1.2)
    assert list(lst) == [1, 1.2, 1.5, 2]

    node12.prepend(1.1)
    assert list(lst) == [1, 1.1, 1.2, 1.5, 2]

    node1.prepend(0)
    assert list(lst) == [0, 1, 1.1, 1.2, 1.5, 2]

    node2.append(3)
    assert list(lst) == [0, 1, 1.1, 1.2, 1.5, 2, 3]

    assert lst.head.data == 0
    assert lst.tail.data == 3

    lst = _HelperBytecodeList([1, 2, 3, 4, 5])
    assert lst.head.data == 1
    assert lst.tail.data == 5



import subprocess
import sys
import json
from _pydev_bundle import pydev_log
import os
import io


def gen_debug_info():
    from _pydevd_bundle.pydevd_constants import DebugInfoHolder
    dct = {}
    for name in (
        'PYDEVD_DEBUG_FILE',
        'DEBUG_TRACE_LEVEL',
        ):
        dct[name] = getattr(DebugInfoHolder, name)

    return dct


if __name__ == "__main__":
    if '-print-debug' in sys.argv:
        info = gen_debug_info()  # break on 2nd process
        pydev_log.info('Something in print-debug')

        print('>>> print-debug pid: %s' % os.getpid())
        print(json.dumps(info))

    else:
        # Note: when running tests we usually have logging setup,
        # so, we create a context so that our changes are restored
        # when it finishes (as the `log_to` function will just reset
        # whatever is there).
        s = io.StringIO()
        with pydev_log.log_context(trace_level=3, stream=s):
            target_log_file = os.getenv('TARGET_LOG_FILE')

            pydev_log.log_to(target_log_file, 1)
            new_debug_info = gen_debug_info()
            subprocess_pid = None
            with subprocess.Popen(
                    [sys.executable, __file__, '-print-debug'],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                ) as process:
                subprocess_pid = process.pid
                stdout, stderr = process.communicate(input)

            output = stdout.decode('utf-8')
            pydev_log.info('Something in initial')

        log_contents = open(pydev_log._compute_filename_with_pid(target_log_file)).read()
        assert 'Something in initial' in log_contents, 'Did not find "Something in initial" in %s' % (log_contents,)

        log_contents = open(pydev_log._compute_filename_with_pid(target_log_file, pid=subprocess_pid)).read()
        assert 'Something in print-debug' in log_contents, 'Did not find "Something in print-debug" in %s' % (log_contents,)

        output = ''.join(output.splitlines(keepends=True)[1:])  # Remove the first line
        loaded_debug_info = json.loads(output)
        assert loaded_debug_info == new_debug_info, 'Expected %s. Found: %s' % (new_debug_info, loaded_debug_info)
        print('>>> Initial pid: %s' % os.getpid())
        print(output)
        print('TEST SUCEEDED')



import sys
from _pydevd_bundle.pydevd_constants import int_types, GENERATED_LEN_ATTR_NAME
from _pydevd_bundle.pydevd_resolver import TOO_LARGE_ATTR
from _pydevd_bundle import pydevd_resolver, pydevd_constants
from _pydevd_bundle import pydevd_frame_utils
import pytest


def get_frame():
    var1 = 1
    var2 = [var1]
    var3 = {33: [var1]}
    return sys._getframe()


def check_vars_dict_expected(as_dict, expected):
    assert as_dict == expected


class _DummyPyDB(object):
    def __init__(self):
        from _pydevd_bundle.pydevd_api import PyDevdAPI

        self.variable_presentation = PyDevdAPI.VariablePresentation()


def test_suspended_frames_manager():
    from _pydevd_bundle.pydevd_suspended_frames import SuspendedFramesManager
    from _pydevd_bundle.pydevd_utils import DAPGrouper

    suspended_frames_manager = SuspendedFramesManager()
    py_db = _DummyPyDB()
    with suspended_frames_manager.track_frames(py_db) as tracker:
        # : :type tracker: _FramesTracker
        thread_id = "thread1"
        frame = get_frame()
        tracker.track(thread_id, pydevd_frame_utils.create_frames_list_from_frame(frame))

        assert suspended_frames_manager.get_thread_id_for_variable_reference(id(frame)) == thread_id

        variable = suspended_frames_manager.get_variable(id(frame))

        # Should be properly sorted.
        assert ["var1", "var2", "var3"] == [x.get_name() for x in variable.get_children_variables()]

        as_dict = dict((x.get_name(), x.get_var_data()) for x in variable.get_children_variables())
        var_reference = as_dict["var2"].pop("variablesReference")
        assert isinstance(var_reference, int_types)  # The variable reference is always a new int.
        assert isinstance(as_dict["var3"].pop("variablesReference"), int_types)  # The variable reference is always a new int.

        check_vars_dict_expected(
            as_dict,
            {
                "var1": {"name": "var1", "value": "1", "type": "int", "evaluateName": "var1", "variablesReference": 0},
                "var2": {"name": "var2", "value": "[1]", "type": "list", "evaluateName": "var2"},
                "var3": {"name": "var3", "value": "{33: [1]}", "type": "dict", "evaluateName": "var3"},
            },
        )

        # Now, same thing with a different format.
        as_dict = dict((x.get_name(), x.get_var_data(fmt={"hex": True})) for x in variable.get_children_variables())
        var_reference = as_dict["var2"].pop("variablesReference")
        assert isinstance(var_reference, int_types)  # The variable reference is always a new int.
        assert isinstance(as_dict["var3"].pop("variablesReference"), int_types)  # The variable reference is always a new int.

        check_vars_dict_expected(
            as_dict,
            {
                "var1": {"name": "var1", "value": "0x1", "type": "int", "evaluateName": "var1", "variablesReference": 0},
                "var2": {"name": "var2", "value": "[0x1]", "type": "list", "evaluateName": "var2"},
                "var3": {"name": "var3", "value": "{0x21: [0x1]}", "type": "dict", "evaluateName": "var3"},
            },
        )

        var2 = dict((x.get_name(), x) for x in variable.get_children_variables())["var2"]
        children_vars = var2.get_children_variables()
        as_dict = dict([x.get_name(), x.get_var_data()] for x in children_vars if x.get_name() not in DAPGrouper.SCOPES_SORTED)
        assert as_dict == {
            "0": {"name": "0", "value": "1", "type": "int", "evaluateName": "var2[0]", "variablesReference": 0},
            GENERATED_LEN_ATTR_NAME: {
                "name": GENERATED_LEN_ATTR_NAME,
                "value": "1",
                "type": "int",
                "evaluateName": "len(var2)",
                "variablesReference": 0,
                "presentationHint": {"attributes": ["readOnly"]},
            },
        }

        var3 = dict((x.get_name(), x) for x in variable.get_children_variables())["var3"]
        children_vars = var3.get_children_variables()
        as_dict = dict([x.get_name(), x.get_var_data()] for x in children_vars if x.get_name() not in DAPGrouper.SCOPES_SORTED)
        assert isinstance(as_dict["33"].pop("variablesReference"), int_types)  # The variable reference is always a new int.

        check_vars_dict_expected(
            as_dict,
            {
                "33": {"name": "33", "value": "[1]", "type": "list", "evaluateName": "var3[33]"},
                GENERATED_LEN_ATTR_NAME: {
                    "name": GENERATED_LEN_ATTR_NAME,
                    "value": "1",
                    "type": "int",
                    "evaluateName": "len(var3)",
                    "variablesReference": 0,
                    "presentationHint": {"attributes": ["readOnly"]},
                },
            },
        )


def get_dict_large_frame():
    obj = {}
    for idx in range(pydevd_constants.PYDEVD_CONTAINER_RANDOM_ACCESS_MAX_ITEMS + +300):
        obj[idx] = 1
    return sys._getframe()


def get_set_large_frame():
    obj = set()
    for idx in range(pydevd_constants.PYDEVD_CONTAINER_RANDOM_ACCESS_MAX_ITEMS + +300):
        obj.add(idx)
    return sys._getframe()


def test_get_child_variables():
    from _pydevd_bundle.pydevd_suspended_frames import SuspendedFramesManager

    suspended_frames_manager = SuspendedFramesManager()
    py_db = _DummyPyDB()
    for frame in (
        get_dict_large_frame(),
        get_set_large_frame(),
    ):
        with suspended_frames_manager.track_frames(py_db) as tracker:
            # : :type tracker: _FramesTracker
            thread_id = "thread1"
            tracker.track(thread_id, pydevd_frame_utils.create_frames_list_from_frame(frame))

            assert suspended_frames_manager.get_thread_id_for_variable_reference(id(frame)) == thread_id

            variable = suspended_frames_manager.get_variable(id(frame))

            children_variables = variable.get_child_variable_named("obj").get_children_variables()

            found_too_large = False
            found_len = False
            for x in children_variables:
                if x.name == TOO_LARGE_ATTR:
                    var_data = x.get_var_data()
                    assert "readOnly" in var_data["presentationHint"]["attributes"]
                    found_too_large = True
                elif x.name == GENERATED_LEN_ATTR_NAME:
                    found_len = True

            if not found_too_large:
                raise AssertionError("Expected to find variable named: %s" % (TOO_LARGE_ATTR,))
            if not found_len:
                raise AssertionError("Expected to find variable named: len()")



from flask import Flask
from flask import render_template

app = Flask(__name__)


@app.route("/")
def home():
    content = "Flask-Jinja-Test"
    return render_template("hello.html", title="Hello", content=content)


@app.route("/handled")
def bad_route_handled():
    try:
        raise ArithmeticError("Hello")
    except Exception:
        pass
    return render_template("hello.html", title="Hello", content="Flask-Jinja-Test")


@app.route("/unhandled")
def bad_route_unhandled():
    raise ArithmeticError("Hello")
    return render_template("hello.html", title="Hello", content="Flask-Jinja-Test")


@app.route("/bad_template")
def bad_template():
    return render_template("bad.html", title="Bad", content="Flask-Jinja-Test")


@app.route("/exit")
def exit_app():
    from flask import request

    func = request.environ.get("werkzeug.server.shutdown")
    if func is None:
        raise RuntimeError("No shutdown")
    func()
    return "Done"


if __name__ == "__main__":
    app.run()



import sys
import os


import unittest


class Test(unittest.TestCase):
    def test_it(self):
        # make it as if we were executing from the directory above this one (so that we can use jycompletionserver
        # without the need for it being in the pythonpath)
        # (twice the dirname to get the previous level from this file.)
        import test_pydevdio  # @UnresolvedImport - importing itself

        ADD_TO_PYTHONPATH = os.path.join(os.path.dirname(os.path.dirname(test_pydevdio.__file__)))
        sys.path.insert(0, ADD_TO_PYTHONPATH)

        try:
            from _pydevd_bundle import pydevd_io

            original = sys.stdout

            try:
                sys.stdout = pydevd_io.IOBuf()
                print("foo")
                print("bar")

                self.assertEqual("foo\nbar\n", sys.stdout.getvalue())  # @UndefinedVariable

                print("ww")
                print("xx")
                self.assertEqual("ww\nxx\n", sys.stdout.getvalue())  # @UndefinedVariable
            finally:
                sys.stdout = original
        finally:
            # remove it to leave it ok for other tests
            sys.path.remove(ADD_TO_PYTHONPATH)



from __future__ import unicode_literals

import os
import subprocess
import sys

miniconda64_envs = os.getenv("MINICONDA64_ENVS")
python_installations = [
    r"%s/py34_64/bin/python" % miniconda64_envs,
    r"%s/py35_64/bin/python" % miniconda64_envs,
    r"%s/py36_64/bin/python" % miniconda64_envs,
    r"%s/py37_64/bin/python" % miniconda64_envs,
]
root_dir = os.path.dirname(os.path.dirname(__file__))


def list_binaries():
    for f in os.listdir(os.path.join(root_dir, "_pydevd_bundle")):
        if f.endswith(".so"):
            yield f


def extract_version(python_install):
    return python_install.split("/")[-3][2:]


def main():
    from generate_code import generate_dont_trace_files
    from generate_code import generate_cython_module

    # First, make sure that our code is up to date.
    generate_dont_trace_files()
    generate_cython_module()

    for python_install in python_installations:
        assert os.path.exists(python_install)

    from build import remove_binaries

    remove_binaries([".so"])

    for f in list_binaries():
        raise AssertionError("Binary not removed: %s" % (f,))

    for i, python_install in enumerate(python_installations):
        new_name = "pydevd_cython_%s_%s" % (sys.platform, extract_version(python_install))
        args = [
            python_install,
            os.path.join(root_dir, "build_tools", "build.py"),
            "--no-remove-binaries",
            "--target-pyd-name=%s" % new_name,
            "--force-cython",
        ]
        if i != 0:
            args.append("--no-regenerate-files")
        version_number = extract_version(python_install)
        if version_number.startswith("36") or version_number.startswith("37"):
            name_frame_eval = "pydevd_frame_evaluator_%s_%s" % (sys.platform, extract_version(python_install))
            args.append("--target-pyd-frame-eval=%s" % name_frame_eval)
        print("Calling: %s" % (" ".join(args)))
        subprocess.check_call(args)


if __name__ == "__main__":
    main()



"""
Helper module to do refactoring to convert names to pep8.
"""

import re
import os
import names_to_rename

_CAMEL_RE = re.compile(r"(?<=[a-z])([A-Z])")
_CAMEL_DEF_RE = re.compile(r"(def )((([A-Z0-9]+|[a-z0-9])[a-z][a-z0-9]*[A-Z]|[a-z0-9]*[A-Z][A-Z0-9]*[a-z])[A-Za-z0-9]*)")


def _normalize(name):
    return _CAMEL_RE.sub(lambda x: "_" + x.group(1).lower(), name).lower()


def find_matches_in_contents(contents):
    return [x[1] for x in re.findall(_CAMEL_DEF_RE, contents)]


def iter_files_in_dir(dirname):
    for root, dirs, files in os.walk(dirname):
        for name in ("pydevd_attach_to_process", ".git", "stubs", "pydev_ipython", "third_party", "pydev_ipython"):
            try:
                dirs.remove(name)
            except:
                pass
        for filename in files:
            if filename.endswith(".py") and filename not in ("rename_pep8.py", "names_to_rename.py"):
                path = os.path.join(root, filename)
                with open(path, "rb") as stream:
                    initial_contents = stream.read()

                yield path, initial_contents


def find_matches():
    found = set()
    for path, initial_contents in iter_files_in_dir(os.path.dirname(os.path.dirname(__file__))):
        found.update(find_matches_in_contents(initial_contents))
    print("\n".join(sorted(found)))
    print("Total", len(found))


def substitute_contents(re_name_to_new_val, initial_contents):
    contents = initial_contents
    for key, val in re_name_to_new_val.iteritems():
        contents = re.sub(key, val, contents)
    return contents


def make_replace():
    re_name_to_new_val = load_re_to_new_val(names_to_rename.NAMES)
    # traverse root directory, and list directories as dirs and files as files
    for path, initial_contents in iter_files_in_dir(os.path.dirname(os.path.dirname(__file__))):
        contents = substitute_contents(re_name_to_new_val, initial_contents)
        if contents != initial_contents:
            print("Changed something at: %s" % (path,))

            for val in re_name_to_new_val.itervalues():
                # Check in initial contents to see if it already existed!
                if re.findall(r"\b%s\b" % (val,), initial_contents):
                    raise AssertionError(
                        "Error in:\n%s\n%s is already being used (and changes may conflict)."
                        % (
                            path,
                            val,
                        )
                    )

            with open(path, "wb") as stream:
                stream.write(contents)


def load_re_to_new_val(names):
    name_to_new_val = {}
    for n in names.splitlines():
        n = n.strip()
        if not n.startswith("#") and n:
            name_to_new_val[r"\b" + n + r"\b"] = _normalize(n)
    return name_to_new_val


def test():
    assert _normalize("RestoreSysSetTraceFunc") == "restore_sys_set_trace_func"
    assert _normalize("restoreSysSetTraceFunc") == "restore_sys_set_trace_func"
    assert _normalize("Restore") == "restore"
    matches = find_matches_in_contents(
        """
    def CamelCase()
    def camelCase()
    def ignore()
    def ignore_this()
    def Camel()
    def CamelCaseAnother()
    """
    )
    assert matches == ["CamelCase", "camelCase", "Camel", "CamelCaseAnother"]
    re_name_to_new_val = load_re_to_new_val(
        """
# Call -- skip
# Call1 -- skip
# Call2 -- skip
# Call3 -- skip
# Call4 -- skip
CustomFramesContainerInit
DictContains
DictItems
DictIterItems
DictIterValues
DictKeys
DictPop
DictValues
"""
    )
    assert re_name_to_new_val == {
        "\\bDictPop\\b": "dict_pop",
        "\\bDictItems\\b": "dict_items",
        "\\bDictIterValues\\b": "dict_iter_values",
        "\\bDictKeys\\b": "dict_keys",
        "\\bDictContains\\b": "dict_contains",
        "\\bDictIterItems\\b": "dict_iter_items",
        "\\bCustomFramesContainerInit\\b": "custom_frames_container_init",
        "\\bDictValues\\b": "dict_values",
    }
    assert (
        substitute_contents(
            re_name_to_new_val,
            """
CustomFramesContainerInit
DictContains
DictItems
DictIterItems
DictIterValues
DictKeys
DictPop
DictValues
""",
        )
        == """
custom_frames_container_init
dict_contains
dict_items
dict_iter_items
dict_iter_values
dict_keys
dict_pop
dict_values
"""
    )


if __name__ == "__main__":
    #     find_matches()
    make_replace()
#     test()
